* eFORTH 1.03  para SINCLAIR QL y compatibles QDOS.
* Ensamblado con el Assembler de Metacomco.
* Escrito por:
*              Salvador Merino
*              CTRA CADIZ, CERAMICAS MARY
*              29640 Torreblanca del Sol
*              Fuengirola (Mlaga)
*              SPAIN
*
* Publicado en:
*              QLIPER (El CLUB ESPAOL de USUARIOS QL)
*              ACACIAS 44
*              28023 MADRID
*              SPAIN
*
*===============================================================
*
*       eForth 1.0 by Bill Muench and C. H. Ting, 1990
*       Much of the code is derived from the following sources:
*              8086 figForth by Thomas Newman, 1981 and Joe smith, 1983
*              aFORTH by John Rible
*              bFORTH by Bill Muench
*
*       The goal of this implementation is to provide a simple eForth Model
*       which can be ported easily to many 8, 16, 24 and 32 bit CPU's.
*       The following attributes make it suitable for CPU's of the '90:
*
*              small machine dependent kernel and portable high level code
*              source code in the MASM format
*              direct threaded code
*              separated code and name dictionaries
*              simple vectored terminal and file interface to host computer
*              aligned with the proposed ANS Forth Standard
*              easy upgrade path to optimize for specific CPU
*
*       You are invited to implement this Model on your favorite CPU and
*       contribute it to the eForth Library for public use. You may use
*       a portable implementation to advertise more sophisticated and
*       optimized version for commercial purposes. However, you are
*       expected to implement the Model faithfully. The eForth Working
*       Group reserves the right to reject implementation which deviates
*       significantly from this Model.
*
*       As the ANS Forth Standard is still evolving, this Model will
*       change accordingly. Implementations must state clearly the
*       version number of the Model being tracked.
*
*       Representing the eForth Working Group in the Silicon Valley FIG Chapter.
*       Send contributions to:
*
*              Dr. C. H. Ting
*              156 14th Avenue
*              San Mateo, CA 94402
*              (415) 571-7639
*
*===============================================================
 
*; Version control
 
VER       equ       1                     *major release version
EXTE      equ       3                     *minor extension
 
*; Constants
 
COMPO     equ       $040                     *lexicon compile only bit
IMEDD     equ       $080                     *lexicon immediate bit
MASKK     equ       $07F1F                     *lexicon bit mask
 
CELLL     equ       4                     *size of a cell
BASEE     equ       10                     *default radix
VOCSS     equ       8                     *depth of vocabulary stack
 
BKSPP     equ       194                    *backspace
LF        equ       10                     *line feed
CRR       equ       10                     *carriage return
ERR       equ       27                     *error escape
TIC       equ       39                     *tick
 
CALLL     equ       $4E714EB9                     *NOP CALL opcodes
 
*; Memory allocation       0//code>--//--<name//up>--<sp//tib>--rp//em
 
EM        equ       262144                    *top of memory
COLDD     equ       229376                     *cold start vector
 
US        equ       64*CELLL              *user area size in cells
RETS      equ       64*CELLL              *return stack/TIB size
 
RPP       equ       EM-8*CELLL              *start of return stack (RP0)
TIBB      equ       RPP-RETS                     *terminal input buffer (TIB)
SPP       equ       TIBB-8*CELLL              *start of data stack (SP0)
UPP       equ       EM-256*CELLL              *start of user area (UP0)
NAMEE     equ       UPP-8*CELLL              *name dictionary
CODEE     equ       COLDD+US              *code dictionary

*; Initialize assembly variables
 
LINK_     SET      0                                   *force a null link
*NAME_     SET      NAMEE                                   *initialize name pointer
CODE_     SET      CODEE                                   *initialize code pointer
USER_     SET      4*CELLL                            *first user variable offset
 
*; Define assembly macros
 
*       Adjust an address to the next cell boundary.
 
ALIGNM    MACRO
          cnop      0,4
          ENDM
 
*       Compile a code definition header.
 
CODEM     MACRO
          dc.l      \3,LINK_
LINK_     SET       *
          dc.b      \1,\2
          ALIGNM
          ENDM
 
*       Assemble inline direct threaded code ending.
 
NEXTM     MACRO
          move.l    (a2)+,a5                              *;next code address into AX
          jmp       (a5)                            *;jump directly to code address
          ENDM
 
*; Main entry points and COLD start data

          ORG       COLDD                                   *beginning of cold boot
          lea.l     RBASIC,a3
          move.l    a7,(a3)
          lea.l     PBLOCK,a1
          move.w    $c6,a2
          jsr       (a2)
          lea.l     CANAL,a3
          move.l    a0,(a3)
          move.l    a0,4(a3)
          move.l    a0,-(a7)

ORIG      lea.l    SPP,a3
          lea.l    RPP,a4
          jmp      COLD                  *to high level cold start


* COLD start moves the following to USER variables.
* MUST BE IN SAME ORDER AS USER VARIABLES.
CANAL     ds.l      1
CDEFECTO  ds.l      1
RBASIC    ds.l      1
PBLOCK    dc.b      4
          dc.b      2
          dc.b      7
          dc.w      512
          dc.w      256
          dc.w      0
          dc.w      0
*
*
          ALIGNM                                          *align to cell boundary
 
UZERO         dc.l       0,0,0,0                 *reserved
              dc.l       SPP                     *SP0
              dc.l       RPP                     *RP0
              dc.l       QRX                     *'?KEY
              dc.l       TXSTO                     *'EMIT
              dc.l       ACCEP                     *'EXPECT
              dc.l       KTAP                     *'TAP
              dc.l       TXSTO                     *'ECHO
              dc.l       DOTOK                     *'PROMPT
              dc.l       BASEE                     *BASE
              dc.l       0                     *tmp
              dc.l       0                     *SPAN
              dc.l       0                     *>IN
              dc.l       0                     *#TIB
              dc.l       TIBB                     *TIB
              dc.l       0                     *CSP
              dc.l       INTER                     *'EVAL
              dc.l       NUMBQ                     *'NUMBER
              dc.l       0                     *HLD
              dc.l       0                     *HANDLER
              dc.l       0               *CONTEXT pointer
              dc.l       0                    *CURRENT pointer
              dc.l       LASTN                 *vocabulary link pointer
              dc.l       CTOP                     *CP
              dc.l       NTOP                     *NP
              dc.l       LASTN                     *LAST
ULAST
 
 
*; Device dependent I/O
 
*   BYE              ( -- )
*              Exit eForth.
          CODEM       3,'BYE',BYE
BYE       move.l    RBASIC,a7
          rts                       *MS-DOS terminate process
 
*   ?RX              ( -- c T | F )
*              Return input character and true, or a false if no input.

          CODEM     3,'?RX',QRX
QRX       moveq     #-1,d3
          movea.l   (a7),a0
          moveq     #$e,d0
          trap      #3
          moveq     #1,d0
          trap      #3
          tst.w     d0
          bne       QRX1
          move.l    d1,-(a3)
          move.l    #-1,-(a3)
          moveq     #$f,d0
          trap      #3
          NEXTM
QRX1
          move.l    (a7)+,a0
          moveq     #2,d0
          trap      #2
          move.l    #0,-(a3)
          NEXTM
*   TX!              ( c -- )
*              Send character c to the output device.

          CODEM     3,'TX!',TXSTO
TXSTO     move.l    (a3)+,d1
          moveq     #-1,d3
          movea.l   CANAL,a0
          moveq     #5,d0
          trap      #3
          NEXTM

*   !IO              ( -- )
*              Initialize the serial I/O devices.

              CODEM       3,'!IO',STOIO
STOIO
              NEXTM
 
*; The kernel
 
*   doLIT       ( -- w )
*              Push an inline literal.

          CODEM     COMPO+5,'doLIT',DOLIT
DOLIT
          move.l    (a2)+,-(a3)
          NEXTM
 
*   doLIST       ( a -- )
*              Process colon list.

          CODEM     COMPO+6,'doLIST',DOLST
DOLST
          move.l    a2,-(a4)
          move.l    (a7)+,a2
          NEXTM
 
*   EXIT       ( -- )
*              Terminate a colon definition.

          CODEM     4,'EXIT',EXIT
EXIT
          move.l    (a4)+,a2
          NEXTM
 
*   EXECUTE       ( ca -- )
*              Execute the word at ca.

          CODEM     7,'EXECUTE',EXECU
EXECU
          move.l    (a3)+,a5
          jmp       (a5)

*   next       ( -- )
*              Run time code for the single index loop.
*              : next ( -- ) \ hilevel model
*                r> r> dup if 1 - >r @ >r exit then drop cell+ >r *

          CODEM     COMPO+4,'next',DONXT
DONXT
          addi.l    #-1,(a4)
          move.l    (a4),d2
          tst.l     d2
          bmi       NEXT1
          move.l    (a2),a2
          NEXTM
NEXT1
          adda.l    #4,a4
          adda.l    #4,a2
          NEXTM
 
*   ?branch       ( f -- )
*              Branch if flag is zero.

          CODEM     COMPO+7,'?branch',QBRAN
QBRAN
          move.l    (a3)+,d4
          tst.l     d4
          beq       BRAN1
          adda.l    #4,a2
          NEXTM
BRAN1     move.l    (a2),a2            *IP:=(IP)
          NEXTM
 
*   branch       ( -- )
*              Branch to an inline address.

          CODEM     COMPO+6,'branch',BRAN
BRAN
          move.l    (a2),a2
          NEXTM
 
*   !              ( w a -- )
*              Pop the data stack to memory.

          CODEM     1,'!',STORE
STORE
          move.l    (a3)+,a5
          move.l    (a3)+,(a5)
          NEXTM
 
*   @              ( a -- w )
*              Push memory location to the data stack.

          CODEM     1,'@',AT
AT
          move.l    (a3)+,a5
          move.l    (a5),-(a3)
          NEXTM
 
*   C!              ( c b -- )
*              Pop the data stack to byte memory.

          CODEM     2,'C!',CSTOR
CSTOR
          move.l    (a3)+,a5
          move.l    (a3)+,d5
          move.b    d5,0(a5)
          NEXTM
 
*   C@              ( b -- c )
*              Push byte memory location to the data stack.

          CODEM     2,'C@',CAT
CAT
          move.l    (a3)+,a5
          clr.l     d5
          move.b    0(a5),d5
          move.l    d5,-(a3)
          NEXTM
 
*   RP@              ( -- a )
*              Push the current RP to the data stack.

          CODEM     3,'RP@',RPAT
RPAT
          move.l    a4,-(a3)
          NEXTM
 
*   RP!              ( a -- )
*              Set the return stack pointer.

          CODEM     COMPO+3,'RP!',RPSTO
RPSTO
          move.l    (a3)+,a4
          NEXTM
 
*   R>              ( -- w )
*              Pop the return stack to the data stack.

          CODEM     2,'R>',RFROM
RFROM
          move.l    (a4)+,-(a3)
          NEXTM
 
*   R@              ( -- w )
*              Copy top of return stack to the data stack.

          CODEM     2,'R@',RAT
RAT
          move.l    (a4),-(a3)
          NEXTM
 
*   >R              ( w -- )
*              Push the data stack to the return stack.

          CODEM     COMPO+2,'>R',TOR
TOR
          move.l    (a3)+,-(a4)
          NEXTM
 
*   SP@              ( -- a )
*              Push the current data stack pointer.

          CODEM     3,'SP@',SPAT
SPAT
          move.l    a3,a5
          move.l    a5,-(a3)
          NEXTM
 
*   SP!              ( a -- )
*              Set the data stack pointer.

          CODEM     3,'SP!',SPSTO
SPSTO
          move.l    (a3)+,a3
          NEXTM
 
*   DROP       ( w -- )
*              Discard top stack item.

          CODEM     4,'DROP',DROP
DROP
          move.l    (a3)+,d4
          NEXTM
 
*   DUP              ( w -- w w )
*              Duplicate the top stack item.

          CODEM     3,'DUP',DUPP
DUPP
          move.l    0(a3),-(a3)
          NEXTM
 
*   SWAP       ( w1 w2 -- w2 w1 )
*              Exchange top two stack items.

          CODEM     4,'SWAP',ESWAP
ESWAP
          move.l    (a3)+,d4
          move.l    (a3)+,d5
          move.l    d4,-(a3)
          move.l    d5,-(a3)
          NEXTM
 
*   OVER       ( w1 w2 -- w1 w2 w1 )
*              Copy second stack item to top.

          CODEM     4,'OVER',OVER
OVER
          move.l    4(a3),-(a3)
          NEXTM
 
*   0<              ( n -- t )
*              Return true if n is negative.

          CODEM     2,'0<',ZLESS
ZLESS
          move.l    (a3)+,d5
          tst.l     d5
          bmi       VERDAD
          move.l    #0,-(a3)
          NEXTM
VERDAD    move.l    #-1,-(a3)
          NEXTM
 
*   AND              ( w w -- w )
*              Bitwise AND.

          CODEM     3,'AND',ANDD
ANDD
          move.l    (a3)+,d4
          move.l    (a3)+,d5
          and.l     d4,d5
          move.l    d5,-(a3)
          NEXTM
 
*   OR              ( w w -- w )
*              Bitwise inclusive OR.

          CODEM     2,'OR',ORR
ORR
          move.l    (a3)+,d4
          move.l    (a3)+,d5
          or.l      d4,d5
          move.l    d5,-(a3)
          NEXTM
 
*   XOR              ( w w -- w )
*              Bitwise exclusive OR.

          CODEM     3,'XOR',XORR
XORR
          move.l    (a3)+,d4
          move.l    (a3)+,d5
          eor.l     d4,d5
          move.l    d5,-(a3)
          NEXTM
 
*   UM+              ( w w -- w cy )
*              Add two numbers, return the sum and carry flag.

          CODEM     3,'UM+',UPLUS
UPLUS
          clr.l     d0
          clr.l     d1
          move      #4,CCR
          move.l    (a3)+,d4
          move.l    (a3)+,d5
          add.l     d4,d5
          addx.l    d0,d1
          move.l    d5,-(a3)
          move.l    d1,-(a3)
          NEXTM
 
*; System and user variables
 
*   doVAR       ( -- a )
*              Run time routine for VARIABLE and CREATE.

              CODEM     COMPO+5,'doVAR',DOVAR
DOVAR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       RFROM,EXIT
 
*   UP              ( -- a )
*              Pointer to the user area.

              CODEM       2,'UP',UP
UP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOVAR
              dc.l       UPP
 
*   doUSER       ( -- a )
*              Run time routine for user variables.

              CODEM     COMPO+6,'doUSER',DOUSE
DOUSE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       RFROM,AT,UP,AT,PLUS,EXIT
 
*   SP0              ( -- a )
*              Pointer to bottom of the data stack.

              CODEM       3,'SP0',SZERO
SZERO
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
 
*   RP0              ( -- a )
*              Pointer to bottom of the return stack.

              CODEM       3,'RP0',RZERO
RZERO
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
 
*   '?KEY       ( -- a )
*              Execution vector of ?KEY.

*              USERM       5,"'?KEY"

          dc.l      TQKEY,LINK_                     *;token pointer and link
LINK_     SET       *                            *;link points to a name string
          dc.b      5,39,'?KEY'               *;name string
          ALIGNM
TQKEY
          nop
          jsr       DOLST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
*   'EMIT       ( -- a )
*              Execution vector of EMIT.

*              USERM       5,"'EMIT"
          dc.l      TEMIT,LINK_                     *;token pointer and link
LINK_     SET       *                            *;link points to a name string
          dc.b      5,39,'EMIT'               *;name string
          ALIGNM
TEMIT
          nop
          jsr       DOLST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
*   'EXPECT       ( -- a )
*              Execution vector of EXPECT.

*              USERM       7,"'EXPECT"
          dc.l      TEXPE,LINK_                     *;token pointer and link
LINK_     SET       *                            *;link points to a name string
          dc.b      7,39,'EXPECT'               *;name string
          ALIGNM
TEXPE
          nop
          jsr       DOLST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
*   'TAP       ( -- a )
*              Execution vector of TAP.

*              USERM       4,"'TAP"
          dc.l       TTAP,LINK_                     *;token pointer and link
LINK_     SET       *                            *;link points to a name string
          dc.b      4,39,'TAP'               *;name string
          ALIGNM
TTAP
          nop
          jsr       DOLST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
*   'ECHO       ( -- a )
*              Execution vector of ECHO.

*              USERM       5,"'ECHO"
          dc.l      TECHO,LINK_                     *;token pointer and link
LINK_     SET       *                            *;link points to a name string
          dc.b      5,39,'ECHO'               *;name string
          ALIGNM
TECHO
          nop
          jsr       DOLST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
*   'PROMPT       ( -- a )
*              Execution vector of PROMPT.

*              USERM       7,"'PROMPT"
          dc.l       TPROM,LINK_                     *;token pointer and link
LINK_     SET       *                            *;link points to a name string
          dc.b      7,'PROMPT'               *;name string
          ALIGNM
TPROM
          nop
          jsr       DOLST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
*   BASE       ( -- a )
*              Storage of the radix base for numeric I/O.

             CODEM       4,'BASE',BASE
BASE
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset

*   tmp              ( -- a )
*              A temporary storage location used in parse and find.

              CODEM       COMPO+3,'tmp',TEMP
TEMP
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
 
*   SPAN       ( -- a )
*              Hold character count received by EXPECT.

              CODEM       4,'SPAN',SPAN
SPAN
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
 
*   >IN              ( -- a )
*              Hold the character pointer while parsing input stream.

              CODEM       3,'>IN',INN
INN
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
 
*   #TIB       ( -- a )
*              Hold the current count and address of the terminal input buffer.

          CODEM       4,'#TIB',NTIB
NTIB
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
USER_     SET       USER_+CELLL
 
*   CSP              ( -- a )
*              Hold the stack pointer for error checking.

              CODEM       3,'CSP',CSP
CSP
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
 
*   'EVAL       ( -- a )
*              Execution vector of EVAL.

*              USERM       5,"'EVAL"
          dc.l       TEVAL,LINK_                     *;token pointer and link
LINK_     SET       *                            *;link points to a name string
          dc.b      5,39,'EVAL'               *;name string
          ALIGNM
TEVAL
          nop
          jsr       DOLST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset

*   'NUMBER       ( -- a )
*              Execution vector of NUMBER?.

*              USERM       7,"'NUMBER"
          dc.l      TNUMB,LINK_                     *;token pointer and link
LINK_     SET       *                            *;link points to a name string
          dc.b      7,39,'NUMBER'               *;name string
          ALIGNM
TNUMB
          nop
          jsr       DOLST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset

*   HLD              ( -- a )
*              Hold a pointer in building a numeric output string.

              CODEM       3,'HLD',HLD
HLD
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
 
*   HANDLER       ( -- a )
*              Hold the return stack pointer for error handling.

              CODEM       7,'HANDLER',HANDL
HANDL
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
 
*   CONTEXT       ( -- a )
*              A area to specify vocabulary search order.

          CODEM     7,'CONTEXT',CNTXT
CNTXT
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
 
*   CURRENT       ( -- a )
*              Point to the vocabulary to be extended.

          CODEM     7,'CURRENT',CRRNT
CRRNT
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
USER_     SET       USER_+CELLL              *vocabulary link pointer
 
*   CP              ( -- a )
*              Point to the top of the code dictionary.

              CODEM       2,'CP',CP
CP
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
 
*   NP              ( -- a )
*              Point to the bottom of the name dictionary.

              CODEM       2,'NP',NP
NP
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset
 
*   LAST       ( -- a )
*              Point to the last name in the name dictionary.

              CODEM       4,'LAST',LAST
LAST
          nop                                     *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      DOUSE,USER_                     *;followed by doUSER and offset
USER_     SET       USER_+CELLL                     *;update user area offset


*; Common functions
 
*   ?DUP       ( w -- w w | 0 )
*              Dup tos if its is not zero.

              CODEM       4,'?DUP',QDUP
QDUP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DUPP
              dc.l       QBRAN,QDUP1
              dc.l       DUPP
QDUP1         dc.l       EXIT
 
*   ROT              ( w1 w2 w3 -- w2 w3 w1 )
*              Rot 3rd item to top.

              CODEM       3,'ROT',ROT
ROT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOR,ESWAP,RFROM,ESWAP,EXIT
 
*   2DROP       ( w w -- )
*              Discard two items on stack.

              CODEM     5,'2DROP',DDROP
DDROP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DROP,DROP,EXIT
 
*   2DUP       ( w1 w2 -- w1 w2 w1 w2 )
*              Duplicate top two items.

              CODEM       4,'2DUP',DDUP
DDUP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       OVER,OVER,EXIT
 
*   +              ( w w -- sum )
*              Add top two items.

              CODEM       1,'+',PLUS
PLUS
          move.l    (a3)+,d2
          move.l    (a3)+,d3
          add.l     d3,d2
          move.l    d2,-(a3)
          NEXTM
 
*   D+              ( d d -- d )
*              Double addition, as an example using UM+.
*
*              COLONM       2,'D+',DPLUS
*              dc.l       TOR,ESWAP,TOR,UPLUS
*              dc.l       RFROM,RFROM,PLUS,PLUS,EXIT
 
*   NOT              ( w -- w )
*              One's complement of tos.

              CODEM       3,'NOT',INVER
INVER
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,-1,XORR,EXIT
 
*   NEGATE       ( n -- -n )
*              Two's complement of tos.

              CODEM       6,'NEGATE',NEGAT
NEGAT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       INVER,DOLIT,1,PLUS,EXIT
 
*   DNEGATE       ( d -- -d )
*              Two's complement of top double.

              CODEM       7,'DNEGATE',DNEGA
DNEGA
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       INVER,TOR,INVER
              dc.l       DOLIT,1,UPLUS
              dc.l       RFROM,PLUS,EXIT
 
*   -              ( n1 n2 -- n1-n2 )
*              Subtraction.

              CODEM       1,'-',SUBB
SUBB
          move.l    (a3)+,d2
          move.l    (a3)+,d1
          sub.l     d2,d1
          move.l    d1,-(a3)
          NEXTM
 
*   ABS              ( n -- n )
*              Return the absolute value of n.

              CODEM       3,'ABS',ABSS
ABSS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DUPP,ZLESS
              dc.l       QBRAN,ABS1
              dc.l       NEGAT
ABS1          dc.l       EXIT
 
*   =              ( w w -- t )
*              Return true if top two are equal.

              CODEM     1,'=',EQUAL
EQUAL
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       XORR
              dc.l       QBRAN,EQU1
              dc.l       DOLIT,0,EXIT              *false flag
EQU1          dc.l       DOLIT,-1,EXIT              *true flag
 
*   U<              ( u u -- t )
*              Unsigned compare of top two items.

              CODEM       2,'U<',ULESS
ULESS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DDUP,XORR,ZLESS
              dc.l       QBRAN,ULES1
              dc.l       ESWAP,DROP,ZLESS,EXIT
ULES1         dc.l       SUBB,ZLESS,EXIT
 
*   <              ( n1 n2 -- t )
*              Signed compare of top two items.

              CODEM       1,'<',LESS
LESS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DDUP,XORR,ZLESS
              dc.l       QBRAN,LESS1
              dc.l       DROP,ZLESS,EXIT
LESS1         dc.l       SUBB,ZLESS,EXIT
 
*   MAX              ( n n -- n )
*              Return the greater of two top stack items.

              CODEM       3,'MAX',MAX
MAX
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DDUP,LESS
              dc.l       QBRAN,MAX1
              dc.l       ESWAP
MAX1          dc.l       DROP,EXIT
 
*   MIN              ( n n -- n )
*              Return the smaller of top two stack items.

              CODEM       3,'MIN',MIN
MIN
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DDUP,ESWAP,LESS
              dc.l       QBRAN,MIN1
              dc.l       ESWAP
MIN1          dc.l       DROP,EXIT
 
*   WITHIN       ( u ul uh -- t )
*              Return true if u is within the range of ul and uh.

              CODEM       6,'WITHIN',WITHI
WITHI
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       OVER,SUBB,TOR                     *ul <= u < uh
              dc.l       SUBB,RFROM,ULESS,EXIT
 
*; Divide
 
*   UM/MOD       ( udl udh u -- ur uq )
*              Unsigned divide of a double by a single. Return mod and quotient.

              CODEM       6,'UM/MOD',UMMOD
UMMOD
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DDUP,ULESS
              dc.l       QBRAN,UMM4
              dc.l       NEGAT,DOLIT,31,TOR
UMM1          dc.l       TOR,DUPP,UPLUS
              dc.l       TOR,TOR,DUPP,UPLUS
              dc.l       RFROM,PLUS,DUPP
              dc.l       RFROM,RAT,ESWAP,TOR
              dc.l       UPLUS,RFROM,ORR
              dc.l       QBRAN,UMM2
              dc.l       TOR,DROP,DOLIT,1,PLUS,RFROM
              dc.l       BRAN,UMM3
UMM2          dc.l       DROP
UMM3          dc.l       RFROM
              dc.l       DONXT,UMM1
              dc.l       DROP,ESWAP,EXIT
UMM4          dc.l       DROP,DDROP
              dc.l       DOLIT,-1,DUPP,EXIT       *overflow, return max
 
*   M/MOD       ( d n -- r q )
*              Signed floored divide of double by single. Return mod and quotient.


          CODEM       5,'M/MOD',MSMOD
MSMOD
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DUPP,ZLESS,DUPP,TOR
              dc.l       QBRAN,MMOD1
              dc.l       NEGAT,TOR,DNEGA,RFROM
MMOD1         dc.l       TOR,DUPP,ZLESS
              dc.l       QBRAN,MMOD2
              dc.l       RAT,PLUS
MMOD2         dc.l       RFROM,UMMOD,RFROM
              dc.l       QBRAN,MMOD3
              dc.l       ESWAP,NEGAT,ESWAP
MMOD3         dc.l       EXIT
 
*   /MOD       ( n n -- r q )
*              Signed divide. Return mod and quotient.

              CODEM       4,'/MOD',SLMOD
SLMOD
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       OVER,ZLESS,ESWAP,MSMOD,EXIT
 
*   MOD              ( n n -- r )
*              Signed divide. Return mod only.

              CODEM       3,'MOD',MODD
MODD
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       SLMOD,DROP,EXIT
 
*   /              ( n n -- q )
*              Signed divide. Return quotient only.

              CODEM       1,'/',SLASH
SLASH
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       SLMOD,ESWAP,DROP,EXIT
 
*; Multiply
 
*   UM*              ( u u -- ud )
*              Unsigned multiply. Return double product.

              CODEM       3,'UM*',UMSTA
UMSTA
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,0,ESWAP,DOLIT,31,TOR
UMST1         dc.l       DUPP,UPLUS,TOR,TOR
              dc.l       DUPP,UPLUS,RFROM,PLUS,RFROM
              dc.l       QBRAN,UMST2
              dc.l       TOR,OVER,UPLUS,RFROM,PLUS
UMST2         dc.l       DONXT,UMST1
              dc.l       ROT,DROP,EXIT
 
*   *              ( n n -- n )
*              Signed multiply. Return single product.

              CODEM       1,'*',STAR
STAR
          nop
          jsr       DOLST
          dc.l      UMSTA,DROP,EXIT
 
*   M*              ( n n -- d )
*              Signed multiply. Return double product.

              CODEM       2,'M*',MSTAR
MSTAR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DDUP,XORR,ZLESS,TOR
              dc.l       ABSS,ESWAP,ABSS,UMSTA
              dc.l       RFROM
              dc.l       QBRAN,MSTA1
              dc.l       DNEGA
MSTA1         dc.l       EXIT
 
*   */MOD       ( n1 n2 n3 -- r q )
*              Multiply n1 and n2, then divide by n3. Return mod and quotient.

              CODEM       5,'*/MOD',SSMOD
SSMOD
          nop
          dc.l      TOR,MSTAR,RFROM,MSMOD,EXIT
 
*   */              ( n1 n2 n3 -- q )
*              Multiply n1 by n2, then divide by n3. Return quotient only.

              CODEM       2,'*/',STASL
STASL
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       SSMOD,ESWAP,DROP,EXIT
 
*; Miscellaneous
 
*   CELL+       ( a -- a )
*              Add cell size in byte to address.

              CODEM       5,'CELL+',CELLP
CELLP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,CELLL,PLUS,EXIT
 
*   CELL-       ( a -- a )
*              Subtract cell size in byte from address.

              CODEM       5,'CELL-',CELLM
CELLM
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,0-CELLL,PLUS,EXIT
 
*   CELLS       ( n -- n )
*              Multiply tos by cell size in bytes.

              CODEM       5,'CELLS',CELLS
CELLS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,CELLL,STAR,EXIT
 
*   ALIGNED       ( b -- a )
*              Align address to the cell boundary.

              CODEM       7,'ALIGNED',ALGND
ALGND
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DUPP,DOLIT,0,DOLIT,CELLL
              dc.l       UMMOD,DROP,DUPP
              dc.l       QBRAN,ALGN1
              dc.l       DOLIT,CELLL,ESWAP,SUBB
ALGN1         dc.l       PLUS,EXIT
 
*   BL              ( -- 32 )
*              Return 32, the blank character.

              CODEM     2,'BL',BLANK
BLANK
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,32,EXIT
 
*   >CHAR       ( c -- c )
*              Filter non-printing characters.

              CODEM     5,'>CHAR',TCHAR
TCHAR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,$07F,ANDD,DUPP       *mask msb
              dc.l       DOLIT,191,BLANK,WITHI       *check for printable
              dc.l       QBRAN,TCHA1
              dc.l       DROP,DOLIT,95              *replace non-printables
TCHA1         dc.l       EXIT
 
*   DEPTH       ( -- n )
*              Return the depth of the data stack.

              CODEM       5,'DEPTH',DEPTH
DEPTH
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       SPAT,SZERO,AT,ESWAP,SUBB
              dc.l       DOLIT,CELLL,SLASH,EXIT
 
*   PICK       ( ... +n -- ... w )
*              Copy the nth stack item to tos.

              CODEM       4,'PICK',PICK
PICK
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,1,PLUS,CELLS
              dc.l       SPAT,PLUS,AT,EXIT
 
*; Memory access
 
*   +!              ( n a -- )
*              Add n to the contents at address a.

              CODEM       2,'+!',PSTOR
PSTOR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       ESWAP,OVER,AT,PLUS
              dc.l       ESWAP,STORE,EXIT
 
*   2!              ( d a -- )
*              Store the double integer to address a.

              CODEM       2,'2!',DSTOR
DSTOR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       ESWAP,OVER,STORE
              dc.l       CELLP,STORE,EXIT
 
*   2@              ( a -- d )
*              Fetch double integer from address a.

              CODEM       2,'2@',DAT
DAT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DUPP,CELLP,AT
              dc.l       ESWAP,AT,EXIT
 
*   COUNT       ( b -- b +n )
*              Return count byte of a string and add 1 to byte address.

              CODEM       5,'COUNT',COUNT
COUNT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DUPP,DOLIT,1,PLUS
              dc.l       ESWAP,CAT,EXIT
 
*   HERE       ( -- a )
*              Return the top of the code dictionary.

              CODEM       4,'HERE',HERE
HERE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       CP,AT,EXIT
 
*   PAD              ( -- a )
*              Return the address of a temporary buffer.

              CODEM       3,'PAD',PAD
PAD
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       HERE,DOLIT,80,PLUS,EXIT
 
*   TIB              ( -- a )
*              Return the address of the terminal input buffer.

              CODEM       3,'TIB',TIB
TIB
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       NTIB,CELLP,AT,EXIT
 
*   @EXECUTE       ( a -- )
*              Execute vector stored in address a.

              CODEM       8,'@EXECUTE',ATEXE
ATEXE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       AT,QDUP                     *?address or zero
              dc.l       QBRAN,EXE1
              dc.l       EXECU                     *execute if non-zero
EXE1          dc.l       EXIT                     *do nothing if zero
 
*   CMOVE       ( b1 b2 u -- )
*              Copy u bytes from b1 to b2.

              CODEM       5,'CMOVE',CMOVE
CMOVE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOR
              dc.l       BRAN,CMOV2
CMOV1         dc.l       TOR,DUPP,CAT
              dc.l       RAT,CSTOR
              dc.l       DOLIT,1,PLUS
              dc.l       RFROM,DOLIT,1,PLUS
CMOV2         dc.l       DONXT,CMOV1
              dc.l       DDROP,EXIT
 
*   FILL       ( b u c -- )
*              Fill u bytes of character c to area beginning at b.

              CODEM       4,'FILL',FILL
FILL
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       ESWAP,TOR,ESWAP
              dc.l       BRAN,FILL2
FILL1         dc.l       DDUP,CSTOR,DOLIT,1,PLUS
FILL2         dc.l       DONXT,FILL1
              dc.l       DDROP,EXIT
 
*   -TRAILING       ( b u -- b u )
*              Adjust the count to eliminate trailing white space.

              CODEM       9,'-TRAILING',DTRAI
DTRAI
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOR
              dc.l       BRAN,DTRA2
DTRA1         dc.l       BLANK,OVER,RAT,PLUS,CAT,LESS
              dc.l       QBRAN,DTRA2
              dc.l       RFROM,DOLIT,1,PLUS,EXIT       *adjusted count
DTRA2         dc.l       DONXT,DTRA1
              dc.l       DOLIT,0,EXIT              *count=0
 
*   PACK$       ( b u a -- a )
*              Build a counted string with u characters from b. Null fill.

              CODEM       5,'PACK$',PACKS
PACKS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       ALGND,DUPP,TOR              *strings only on cell boundary
              dc.l       OVER,DUPP,DOLIT,0
              dc.l       DOLIT,CELLL,UMMOD,DROP       *count mod cell
              dc.l       SUBB,OVER,PLUS
              dc.l       DOLIT,0,ESWAP,STORE       *null fill cell
              dc.l       DDUP,CSTOR,DOLIT,1,PLUS       *save count
              dc.l       ESWAP,CMOVE,RFROM,EXIT       *move string
 
*; Numeric output, single precision
 
*   DIGIT       ( u -- c )
*              Convert digit u to a character.

              CODEM       5,'DIGIT',DIGIT
DIGIT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,9,OVER,LESS
              dc.l       DOLIT,7,ANDD,PLUS
              dc.l       DOLIT,48,PLUS,EXIT
 
*   EXTRACT       ( n base -- n c )
*              Extract the least significant digit from n.

              CODEM       7,'EXTRACT',EXTRC
EXTRC
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,0,ESWAP,UMMOD
              dc.l       ESWAP,DIGIT,EXIT
 
*   <#              ( -- )
*              Initiate the numeric output process.

              CODEM       2,'<#',BDIGS
BDIGS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       PAD,HLD,STORE,EXIT
 
*   HOLD       ( c -- )
*              Insert a character into the numeric output string.

              CODEM       4,'HOLD',HOLD
HOLD
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       HLD,AT,DOLIT,1,SUBB
              dc.l       DUPP,HLD,STORE,CSTOR,EXIT
 
*   #              ( u -- u )
*              Extract one digit from u and append the digit to output string.

              CODEM       1,'#',DIG
DIG
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       BASE,AT,EXTRC,HOLD,EXIT
 
*   #S              ( u -- 0 )
*              Convert u until all digits are added to the output string.

              CODEM       2,'#S',DIGS
DIGS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
DIGS1         dc.l       DIG,DUPP
              dc.l       QBRAN,DIGS2
              dc.l       BRAN,DIGS1
DIGS2         dc.l       EXIT
 
*   SIGN       ( n -- )
*              Add a minus sign to the numeric output string.

              CODEM       4,'SIGN',SIGN
SIGN
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       ZLESS
              dc.l       QBRAN,SIGN1
              dc.l       DOLIT,45,HOLD
SIGN1         dc.l       EXIT
 
*   #>              ( w -- b u )
*              Prepare the output string to be TYPE'd.

              CODEM       2,'#>',EDIGS
EDIGS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DROP,HLD,AT
              dc.l       PAD,OVER,SUBB,EXIT
 
*   str              ( n -- b u )
*              Convert a signed integer to a numeric string.

              CODEM       3,'str',STR
STR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DUPP,TOR,ABSS
              dc.l       BDIGS,DIGS,RFROM
              dc.l       SIGN,EDIGS,EXIT
 
*   HEX              ( -- )
*              Use radix 16 as base for numeric conversions.

              CODEM       3,'HEX',HEX
HEX
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,16,BASE,STORE,EXIT
 
*   DECIMAL       ( -- )
*              Use radix 10 as base for numeric conversions.

              CODEM       7,'DECIMAL',DECIM
DECIM
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,10,BASE,STORE,EXIT
 
*; Numeric input, single precision
 
*   DIGIT?       ( c base -- u t )
*              Convert a character to its numeric value. A flag indicates success.

              CODEM       6,'DIGIT?',DIGTQ
DIGTQ
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOR,DOLIT,48,SUBB
              dc.l       DOLIT,9,OVER,LESS
              dc.l       QBRAN,DGTQ1
              dc.l       DOLIT,7,SUBB
              dc.l       DUPP,DOLIT,10,LESS,ORR
DGTQ1         dc.l       DUPP,RFROM,ULESS,EXIT
 
*   NUMBER?       ( a -- n T | a F )
*              Convert a number string to integer. Push a flag on tos.

              CODEM       7,'NUMBER?',NUMBQ
NUMBQ
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       BASE,AT,TOR,DOLIT,0,OVER,COUNT
              dc.l       OVER,CAT,DOLIT,36,EQUAL
              dc.l       QBRAN,NUMQ1
              dc.l       HEX,ESWAP,DOLIT,1,PLUS
              dc.l       ESWAP,DOLIT,1,SUBB
NUMQ1         dc.l       OVER,CAT,DOLIT,45,EQUAL,TOR
              dc.l       ESWAP,RAT,SUBB,ESWAP,RAT,PLUS,QDUP
              dc.l       QBRAN,NUMQ6
              dc.l       DOLIT,1,SUBB,TOR
NUMQ2         dc.l       DUPP,TOR,CAT,BASE,AT,DIGTQ
              dc.l       QBRAN,NUMQ4
              dc.l       ESWAP,BASE,AT,STAR,PLUS,RFROM
              dc.l       DOLIT,1,PLUS
              dc.l       DONXT,NUMQ2
              dc.l       RAT,ESWAP,DROP
              dc.l       QBRAN,NUMQ3
              dc.l       NEGAT
NUMQ3         dc.l       ESWAP
              dc.l       BRAN,NUMQ5
NUMQ4         dc.l       RFROM,RFROM,DDROP,DDROP,DOLIT,0
NUMQ5         dc.l       DUPP
NUMQ6         dc.l       RFROM,DDROP
              dc.l       RFROM,BASE,STORE,EXIT
 
*; Basic I/O

*   ?KEY       ( -- c T | F )
*              Return input character and true, or a false if no input.

              CODEM       4,'?KEY',QKEY
QKEY
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TQKEY,ATEXE,EXIT
 
*   KEY              ( -- c )
*              Wait for and return an input character.

              CODEM       3,'KEY',KEY
KEY
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
KEY1          dc.l       QKEY
              dc.l       QBRAN,KEY1
              dc.l       EXIT
 
*   EMIT       ( c -- )
*              Send a character to the output device.

              CODEM       4,'EMIT',EMIT
EMIT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TEMIT,ATEXE,EXIT
 
*   NUF?       ( -- t )
*              Return false if no input, else pause and if CR return true.

              CODEM       4,'NUF?',NUFQ
NUFQ
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       QKEY,DUPP
              dc.l       QBRAN,NUFQ1
              dc.l       DDROP,KEY,DOLIT,CRR,EQUAL
NUFQ1         dc.l       EXIT
 
*   PACE       ( -- )
*              Send a pace character for the file downloading process.

              CODEM       4,'PACE',PACE
PACE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,11,EMIT,EXIT
 
*   SPACE       ( -- )
*              Send the blank character to the output device.

              CODEM       5,'SPACE',SPACE
SPACE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       BLANK,EMIT,EXIT
 
*   SPACES       ( +n -- )
*              Send n spaces to the output device.

              CODEM       6,'SPACES',SPACS
SPACS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,0,MAX,TOR
              dc.l       BRAN,CHAR2
CHAR1         dc.l       SPACE
CHAR2         dc.l       DONXT,CHAR1
              dc.l       EXIT
 
*   TYPE       ( b u -- )
*              Output u characters from b.

              CODEM       4,'TYPE',TYPEE
TYPEE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOR
              dc.l       BRAN,TYPE2
TYPE1         dc.l       DUPP,CAT,EMIT
              dc.l       DOLIT,1,PLUS
TYPE2         dc.l       DONXT,TYPE1
              dc.l       DROP,EXIT
 
*   CR              ( -- )
*              Output a carriage return and a line feed.

              CODEM       2,'CR',CR
CR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,CRR,EMIT
              dc.l       EXIT
 
*   do$              ( -- a )
*              Return the address of a compiled string.

              CODEM       COMPO+3,'do$',DOSTR
DOSTR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       RFROM,RAT,RFROM,COUNT,PLUS
              dc.l       ALGND,TOR,ESWAP,TOR,EXIT
 
*   $"|              ( -- a )
*              Run time routine compiled by $". Return address of a compiled string.

              CODEM       COMPO+3,'$"|',STRQP
STRQP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOSTR,EXIT              *force a call to do$
 
*   ."|              ( -- )
*              Run time routine of ." . Output a compiled string.

              CODEM     COMPO+3,'."|',DOTQP
DOTQP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOSTR,COUNT,TYPEE,EXIT
 
*   .R              ( n +n -- )
*              Display an integer in a field of n columns, right justified.

              CODEM       2,'.R',DOTR
DOTR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOR,STR,RFROM,OVER,SUBB
              dc.l       SPACS,TYPEE,EXIT
 
*   U.R              ( u +n -- )
*              Display an unsigned integer in n column, right justified.

              CODEM       3,'U.R',UDOTR
UDOTR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOR,BDIGS,DIGS,EDIGS
              dc.l       RFROM,OVER,SUBB
              dc.l       SPACS,TYPEE,EXIT
 
*   U.              ( u -- )
*              Display an unsigned integer in free format.

              CODEM       2,'U.',UDOT
UDOT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       BDIGS,DIGS,EDIGS
              dc.l       SPACE,TYPEE,EXIT
 
*   .              ( w -- )
*              Display an integer in free format, preceeded by a space.

              CODEM       1,'.',DOT
DOT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       BASE,AT,DOLIT,10,XORR       *?decimal
              dc.l       QBRAN,DOT1
              dc.l       UDOT,EXIT              *no, display unsigned
DOT1          dc.l       STR,SPACE,TYPEE,EXIT       *yes, display signed
 
*   ?              ( a -- )
*              Display the contents in a memory cell.

              CODEM       1,'?',QUEST
QUEST
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       AT,DOT,EXIT
 
*; Parsing
 
*   parse       ( b u c -- b u delta * <string> )
*              Scan string delimited by c. Return found string and its offset.

              CODEM       5,'parse',PARS
PARS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TEMP,STORE,OVER,TOR,DUPP
              dc.l       QBRAN,PARS8
              dc.l       DOLIT,1,SUBB,TEMP,AT,BLANK,EQUAL
              dc.l       QBRAN,PARS3
              dc.l       TOR
PARS1         dc.l       BLANK,OVER,CAT              *skip leading blanks ONLY
              dc.l       SUBB,ZLESS,INVER
              dc.l       QBRAN,PARS2
              dc.l       DOLIT,1,PLUS
              dc.l       DONXT,PARS1
              dc.l       RFROM,DROP,DOLIT,0,DUPP,EXIT
PARS2         dc.l       RFROM
PARS3         dc.l       OVER,ESWAP
              dc.l       TOR
PARS4         dc.l       TEMP,AT,OVER,CAT,SUBB       *scan for delimiter
              dc.l       TEMP,AT,BLANK,EQUAL
              dc.l       QBRAN,PARS5
              dc.l       ZLESS
PARS5         dc.l       QBRAN,PARS6
              dc.l       DOLIT,1,PLUS
              dc.l       DONXT,PARS4
              dc.l       DUPP,TOR
              dc.l       BRAN,PARS7
PARS6         dc.l       RFROM,DROP,DUPP
              dc.l       DOLIT,1,PLUS,TOR
PARS7         dc.l       OVER,SUBB
              dc.l       RFROM,RFROM,SUBB,EXIT
PARS8         dc.l       OVER,RFROM,SUBB,EXIT
 
*   PARSE       ( c -- b u * <string> )
*              Scan input stream and return counted string delimited by c.

              CODEM       5,'PARSE',PARSE
PARSE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOR,TIB,INN,AT,PLUS       *current input buffer pointer
              dc.l       NTIB,AT,INN,AT,SUBB       *remaining count
              dc.l       RFROM,PARS,INN,PSTOR,EXIT
 
*   .(              ( -- )
*              Output following string up to next ) .

              CODEM       IMEDD+2,'.(',DOTPR
DOTPR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,41,PARSE,TYPEE,EXIT
 
*   (              ( -- )
*              Ignore following string up to next ) . A comment.

              CODEM       IMEDD+1,40,PAREN
PAREN
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,41,PARSE,DDROP,EXIT
 
*   \              ( -- )
*              Ignore following text till the end of line.

              CODEM       IMEDD+1,'\',BKSLA
BKSLA
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       NTIB,AT,INN,STORE,EXIT
 
*   CHAR       ( -- c )
*              Parse next word and return its first character.

              CODEM     4,'CHAR',CHAR
CHAR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       BLANK,PARSE,DROP,CAT,EXIT
 
*   TOKEN       ( -- a * <string> )
*              Parse a word from input stream and copy it to name dictionary.

              CODEM       5,'TOKEN',TOKEN
TOKEN
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       BLANK,PARSE,DOLIT,31,MIN
              dc.l       NP,AT,OVER,SUBB,CELLM
              dc.l       PACKS,EXIT
 
*   WORD       ( c -- a * <string> )
*              Parse a word from input stream and copy it to code dictionary.

              CODEM       4,'WORD',WORDD
WORDD
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       PARSE,HERE,PACKS,EXIT
 
*; Dictionary search
 
*   NAME>       ( na -- ca )
*              Return a code address given a name address.

              CODEM       5,'NAME>',NAMET
NAMET
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       CELLM,CELLM,AT,EXIT
 
*   SAME?       ( a a u -- a a f \ -0+ )
*              Compare u cells in two strings. Return 0 if identical.

              CODEM       5,'SAME?',SAMEQ
SAMEQ
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOR
              dc.l       BRAN,SAME2
SAME1         dc.l       OVER,RAT,CELLS,PLUS,AT
              dc.l       OVER,RAT,CELLS,PLUS,AT
              dc.l       SUBB,QDUP
              dc.l       QBRAN,SAME2
              dc.l       RFROM,DROP,EXIT              *strings not equal
SAME2         dc.l       DONXT,SAME1
              dc.l       DOLIT,0,EXIT              *strings equal
 
*   find       ( a va -- ca na | a F )
*              Search a vocabulary for a string. Return ca and na if succeeded.

              CODEM       4,'find',FIND
FIND
          move.l    (a3)+,-(a4)
FIND3
          move.l    (a4),a5
          move.b    (a5)+,d0
          and.b     #$7F1F,d0
          move.l    0(a3),a1
          move.b    (a1)+,d1
          cmp.b     d0,d1
          bne       FIND1
          clr.l     d4
          move.b    d1,d4
          addi.l    #-1,d4
FIND2
          cmpm.b    (a1)+,(a5)+
          bne       FIND1
          dbf       d4,FIND2
          move.l    (a4)+,a5
          move.l    a5,-(a4)
          move.l    -(a5),(a3)
          move.l    -(a5),(a3)
          move.l    (a4)+,-(a3)
          NEXTM
FIND1
          move.l    (a4)+,a5
          suba.l    #4,a5
          tst.l     (a5)
          beq       FIND4
          move.l    (a5),-(a4)
          bra       FIND3
FIND4
          move.l    #0,-(a3)
          NEXTM
 
*   NAME?       ( a -- ca na | a F )
*              Search all context vocabularies for a string.

              CODEM       5,'NAME?',NAMEQ
NAMEQ
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       CNTXT,AT,AT,FIND,EXIT
 
*; Terminal response
SD_PCOL
          moveq     #-1,d3
          move.l    CANAL,a0
          moveq     #$13,d0
          trap      #3
          NEXTM

*   ^H              ( bot eot cur -- bot eot cur )
*              Backup the cursor by one character.

              CODEM       2,'^H',BKSP
BKSP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOR,OVER,RFROM,ESWAP,OVER,XORR
              dc.l       QBRAN,BACK1
              dc.l       SD_PCOL,DOLIT,1,SUBB
              dc.l       BLANK,TECHO,ATEXE
              dc.l       SD_PCOL
BACK1         dc.l       EXIT
 
*   TAP              ( bot eot cur c -- bot eot cur )
*              Accept and echo the key stroke and bump the cursor.

              CODEM       3,'TAP',TAP
TAP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DUPP,TECHO,ATEXE
              dc.l       OVER,CSTOR,DOLIT,1,PLUS,EXIT
 
*   kTAP       ( bot eot cur c -- bot eot cur )
*              Process a key stroke, CR or backspace.

              CODEM       4,'kTAP',KTAP
KTAP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DUPP,DOLIT,CRR,XORR
              dc.l       QBRAN,KTAP2
              dc.l       DOLIT,BKSPP,XORR
              dc.l       QBRAN,KTAP1
              dc.l       BLANK,TAP,EXIT
KTAP1         dc.l       BKSP,EXIT
KTAP2         dc.l       DROP,ESWAP,DROP,DUPP,EXIT
 
*   accept       ( b u -- b u )
*              Accept characters to input buffer. Return with actual count.

              CODEM       6,'accept',ACCEP
ACCEP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       OVER,PLUS,OVER
ACCP1         dc.l       DDUP,XORR
              dc.l       QBRAN,ACCP4
              dc.l       KEY,DUPP
*             dc.l       BLANK,SUBB,DOLIT,95,ULESS
              dc.l       BLANK,DOLIT,127,WITHI
              dc.l       QBRAN,ACCP2
              dc.l       TAP
              dc.l       BRAN,ACCP3
ACCP2         dc.l       TTAP,ATEXE
ACCP3         dc.l       BRAN,ACCP1
ACCP4         dc.l       DROP,OVER,SUBB,EXIT
 
*   EXPECT       ( b u -- )
*              Accept input stream and store count in SPAN.

              CODEM       6,'EXPECT',EXPEC
EXPEC
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TEXPE,ATEXE,SPAN,STORE,DROP,EXIT
 
*   QUERY       ( -- )
*              Accept input stream to terminal input buffer.

              CODEM     5,'QUERY',QUERY
QUERY
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TIB,DOLIT,80,TEXPE,ATEXE,NTIB,STORE
              dc.l       DROP,DOLIT,0,INN,STORE,EXIT
 
*; Error handling
 
*   CATCH       ( ca -- 0 | err# )
*              Execute word at ca and set up an error frame for it.

              CODEM       5,'CATCH',CATCH
CATCH
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       SPAT,TOR,HANDL,AT,TOR       *save error frame
              dc.l       RPAT,HANDL,STORE,EXECU       *execute
              dc.l       RFROM,HANDL,STORE       *restore error frame
              dc.l       RFROM,DROP,DOLIT,0,EXIT       *no error
 
*   THROW       ( err# -- err# )
*              Reset system to current local error frame an update error flag.

              CODEM       5,'THROW',THROW
THROW
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       HANDL,AT,RPSTO              *restore return stack
              dc.l       RFROM,HANDL,STORE       *restore handler frame
              dc.l       RFROM,ESWAP,TOR,SPSTO       *restore data stack
              dc.l       DROP,RFROM,EXIT


*   NULL$       ( -- a )
*              Return address of a null string with zero count.

              CODEM     5,'NULL$',NULLS
NULLS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOVAR                     *emulate CREATE
              dc.l       0
          dc.b      'coyote'
              ALIGNM
 
*   ABORT       ( -- )
*              Reset data stack and jump to QUIT.

              CODEM       5,'ABORT',ABORT
ABORT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       NULLS,THROW
 
*   abort"       ( f -- )
*              Run time routine of ABORT" . Abort with a message.

              CODEM       COMPO+6,'abort"',ABORQ
ABORQ
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       QBRAN,ABOR1              *text flag
              dc.l       DOSTR,THROW              *pass error string
ABOR1         dc.l       DOSTR,DROP,EXIT              *drop error
 
*; The text interpreter
 
*   $INTERPRET       ( a -- )
*              Interpret a word. If failed, try to convert it to an integer.

              CODEM       10,'$INTERPRET',INTER
INTER
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       NAMEQ,QDUP              *?defined
              dc.l       QBRAN,INTE1
              dc.l       AT,DOLIT,$40000000,ANDD       *?compile only lexicon bits
*          DMM       ABORQ,' compile only '
          dc.l      ABORQ
          dc.b      13,' compile only'
          ALIGNM
              dc.l       EXECU,EXIT              *execute defined word
INTE1         dc.l       TNUMB,ATEXE              *convert a number
              dc.l       QBRAN,INTE2
              dc.l       EXIT
INTE2         dc.l       THROW                     *error
 
*   [              ( -- )
*              Start the text interpreter.

              CODEM       IMEDD+1,'[',LBRAC
LBRAC
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,INTER,TEVAL,STORE,EXIT
 
*   .OK              ( -- )
*              Display 'ok' only while interpreting.

              CODEM       3,'.OK',DOTOK
DOTOK
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,INTER,TEVAL,AT,EQUAL
              dc.l       QBRAN,DOTO1
*          DMM       DOTQP,' ok '
          dc.l      DOTQP
          dc.b      3,' ok'
          ALIGNM
DOTO1         dc.l       CR,EXIT
 
*   ?STACK       ( -- )
*              Abort if the data stack underflows.

              CODEM       6,'?STACK',QSTAC
QSTAC
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DEPTH,ZLESS              *check only for underflow
*          DMM       ABORQ,' underflow'
          dc.l      ABORQ
          dc.b      10,' underflow'
          ALIGNM
              dc.l       EXIT
 
*   EVAL       ( -- )
*              Interpret the input stream.

              CODEM       4,'EVAL',EVAL
EVAL
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
EVAL1         dc.l       TOKEN,DUPP,CAT              *?input stream empty
              dc.l       QBRAN,EVAL2
              dc.l       TEVAL,ATEXE,QSTAC       *evaluate input, check stack
              dc.l       BRAN,EVAL1
EVAL2         dc.l       DROP,TPROM,ATEXE,EXIT       *prompt
 
*; Shell
 
*   PRESET       ( -- )
*              Reset data stack pointer and the terminal input buffer.

              CODEM       6,'PRESET',PRESE
PRESE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       SZERO,AT,SPSTO
              dc.l       DOLIT,TIBB,NTIB,CELLP,STORE,EXIT
 
*   xio              ( a a a -- )
*              Reset the I/O vectors 'EXPECT, 'TAP, 'ECHO and 'PROMPT.

              CODEM       COMPO+3,'xio',XIO
XIO
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,ACCEP,TEXPE,DSTOR
              dc.l       TECHO,DSTOR,EXIT
 
*   FILE       ( -- )
*              Select I/O vectors for file download.

              CODEM       4,'FILE',FILE
FILE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,PACE,DOLIT,DROP
              dc.l       DOLIT,KTAP,XIO,EXIT
 
*   HAND       ( -- )
*              Select I/O vectors for terminal interface.

              CODEM       4,'HAND',HAND
HAND
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,DOTOK,DOLIT,EMIT
              dc.l       DOLIT,KTAP,XIO,EXIT
 
*   I/O              ( -- a )
*              Array to store default I/O vectors.

              CODEM       3,'I/O',ISLO
ISLO
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOVAR                     *emulate CREATE
              dc.l       QRX,TXSTO              *default I/O vectors
 
*   CONSOLE       ( -- )
*              Initiate terminal interface.

              CODEM       7,'CONSOLE',CONSO
CONSO
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       ISLO,DAT,TQKEY,DSTOR       *restore default I/O device
              dc.l       HAND,EXIT              *keyboard input
 
*   QUIT       ( -- )
*              Reset return stack pointer and start text interpreter.

              CODEM       4,'QUIT',QUIT
QUIT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       RZERO,AT,RPSTO              *reset return stack pointer
QUIT1         dc.l       LBRAC                     *start interpretation
QUIT2         dc.l       QUERY                     *get input
              dc.l       DOLIT,EVAL,CATCH,QDUP       *evaluate input
              dc.l       QBRAN,QUIT2              *continue till error
              dc.l       TPROM,AT,ESWAP              *save input device
              dc.l       CONSO,NULLS,OVER,XORR       *?display error message
              dc.l       QBRAN,QUIT3
              dc.l       SPACE,COUNT,TYPEE       *error message
*          DMM       DOTQP,' ?  '         *error prompt
          dc.l      DOTQP
          dc.b      3,' ? '
          ALIGNM
QUIT3         dc.l       DOLIT,DOTOK,XORR       *?file input
              dc.l       QBRAN,QUIT4
              dc.l       DOLIT,ERR,EMIT              *file error, tell host
QUIT4         dc.l       PRESE                     *some cleanup
              dc.l       BRAN,QUIT1
 
*; The compiler
 
*   '              ( -- ca )
*              Search context vocabularies for the next word in input stream.

              CODEM     1,39,TICK
TICK
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOKEN,NAMEQ              *?defined
              dc.l       QBRAN,TICK1
              dc.l       EXIT                     *yes, push code address
TICK1         dc.l       THROW                     *no, error
 
*   ALLOT       ( n -- )
*              Allocate n bytes to the code dictionary.

              CODEM     5,'ALLOT',ALLOT
ALLOT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       CP,PSTOR,EXIT              *adjust code pointer
 
*   ,              ( w -- )
*              Compile an integer into the code dictionary.

              CODEM     1,44,COMMA
COMMA
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       HERE,DUPP,CELLP              *cell boundary
              dc.l       CP,STORE,STORE,EXIT       *adjust code pointer, compile
 
*   [COMPILE]       ( -- * <string> )
*              Compile the next immediate word into code dictionary.

              CODEM     IMEDD+9,'[COMPILE]',BCOMP
BCOMP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TICK,COMMA,EXIT
 
*   COMPILE       ( -- )
*              Compile the next address in colon list to code dictionary.

              CODEM     COMPO+7,'COMPILE',COMPI
COMPI
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       RFROM,DUPP,AT,COMMA       *compile address
              dc.l       CELLP,TOR,EXIT              *adjust return address
 
*   LITERAL       ( w -- )
*              Compile tos to code dictionary as an integer literal.

              CODEM       IMEDD+7,'LITERAL',LITER
LITER
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       COMPI,DOLIT,COMMA,EXIT
 
*   $,"              ( -- )
*              Compile a literal string up to next " .

*              COLONM       3,'$,"',STRCQ

          dc.l      STRCQ,LINK_                     *;token pointer and link
LINK_     SET       *                            *;link points to a name string
          dc.b      3,36,44,34               *;name string
          ALIGNM
STRCQ
          nop
          jsr       DOLST
              dc.l       DOLIT,34,WORDD              *move string to code dictionary
              dc.l       COUNT,PLUS,ALGND       *calculate aligned end of string
              dc.l       CP,STORE,EXIT              *adjust the code pointer
 
*   RECURSE       ( -- )
*              Make the current word available for compilation.

              CODEM       IMEDD+7,'RECURSE',RECUR
RECUR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       LAST,AT,NAMET,COMMA,EXIT
 
*; Structures
 
*   FOR              ( -- a )
*              Start a FOR-NEXT loop structure in a colon definition.

              CODEM       IMEDD+3,'FOR',FOR
FOR
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       COMPI,TOR,HERE,EXIT
 
*   BEGIN       ( -- a )
*              Start an infinite or indefinite loop structure.

              CODEM       IMEDD+5,'BEGIN',BEGIN
BEGIN
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       HERE,EXIT
 
*   NEXT       ( a -- )
*              Terminate a FOR-NEXT loop structure.

              CODEM       IMEDD+4,'NEXT',NEXT
NEXT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       COMPI,DONXT,COMMA,EXIT
 
*   UNTIL       ( a -- )
*              Terminate a BEGIN-UNTIL indefinite loop structure.

              CODEM       IMEDD+5,'UNTIL',UNTIL
UNTIL
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       COMPI,QBRAN,COMMA,EXIT
 
*   AGAIN       ( a -- )
*              Terminate a BEGIN-AGAIN infinite loop structure.

              CODEM       IMEDD+5,'AGAIN',AGAIN
AGAIN
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       COMPI,BRAN,COMMA,EXIT
 
*   IF              ( -- A )
*              Begin a conditional branch structure.

              CODEM       IMEDD+2,'IF',IFF
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
IFF
              dc.l       COMPI,QBRAN,HERE
              dc.l       DOLIT,0,COMMA,EXIT
 
*   AHEAD       ( -- A )
*              Compile a forward branch instruction.

              CODEM       IMEDD+5,'AHEAD',AHEAD
AHEAD
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       COMPI,BRAN,HERE,DOLIT,0,COMMA,EXIT
 
*   REPEAT       ( A a -- )
*              Terminate a BEGIN-WHILE-REPEAT indefinite loop.

              CODEM       IMEDD+6,'REPEAT',REPEA
REPEA
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       AGAIN,HERE,ESWAP,STORE,EXIT
 
*   THEN       ( A -- )
*              Terminate a conditional branch structure.

              CODEM       IMEDD+4,'THEN',THENN
THENN
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       HERE,ESWAP,STORE,EXIT
 
*   AFT              ( a -- a A )
*              Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.

              CODEM       IMEDD+3,'AFT',AFT
AFT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DROP,AHEAD,BEGIN,ESWAP,EXIT
 
*   ELSE       ( A -- A )
*              Start the false clause in an IF-ELSE-THEN structure.

              CODEM       IMEDD+4,'ELSE',ELSEE
ELSEE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       AHEAD,ESWAP,THENN,EXIT
 
*   WHILE       ( a -- A a )
*              Conditional branch out of a BEGIN-WHILE-REPEAT loop.

              CODEM       IMEDD+5,'WHILE',WHILE
WHILE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       IFF,ESWAP,EXIT
 
*   ABORT"       ( -- * <string> )
*              Conditional abort with an error message.

              CODEM       IMEDD+6,'ABORT"',ABRTQ
ABRTQ
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       COMPI,ABORQ,STRCQ,EXIT
 
*   $"              ( -- * <string> )
*              Compile an inline string literal.

              CODEM       IMEDD+2,'$"',STRQ
STRQ
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       COMPI,STRQP,STRCQ,EXIT
 
*   ."              ( -- * <string> )
*              Compile an inline string literal to be typed out at run time.

              CODEM       IMEDD+2,'."',DOTQ
DOTQ
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       COMPI,DOTQP,STRCQ,EXIT
 
*; Name compiler
 
*   ?UNIQUE       ( a -- a )
*              Display a warning message if the word already exists.

              CODEM       7,'?UNIQUE',UNIQU
UNIQU
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DUPP,NAMEQ              *?name exists
              dc.l       QBRAN,UNIQ1              *redefinitions are OK
*          DMM       DOTQP,' reDef  '     *but warn the user
          dc.l      DOTQP
          dc.b      7,' reDef '
          ALIGNM
              dc.l       OVER,COUNT,TYPEE       *just in case its not planned
UNIQ1         dc.l       DROP,EXIT
 
*   $,n              ( na -- )
*              Build a new dictionary name using the string at na.

*              COLONM     3,'$,n'
          dc.l      SNAME,LINK_                     *;token pointer and link
LINK_     SET       *                            *;link points to a name string
          dc.b      3,36,44,110               *;name string
          ALIGNM
SNAME
          nop
          jsr       DOLST
              dc.l       DUPP,CAT              *?null input
              dc.l       QBRAN,PNAM1
              dc.l       UNIQU                     *?redefinition
              dc.l       DUPP,LAST,STORE              *save na for vocabulary link
              dc.l       HERE,ALGND,ESWAP              *align code address
              dc.l       CELLM                     *link address
              dc.l       CRRNT,AT,AT,OVER,STORE
              dc.l       CELLM,DUPP,NP,STORE       *adjust name pointer
              dc.l       STORE,EXIT              *save code pointer
PNAM1
          dc.l      STRQP
          dc.b      6,' name '
          ALIGNM
*     DMM       STRQP,' name '       *null input
          dc.l       THROW
 
*; FORTH compiler
 
*   $COMPILE       ( a -- )
*              Compile next word to code dictionary as a token or literal.

              CODEM       8,'$COMPILE',SCOMP
SCOMP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       NAMEQ,QDUP              *?defined
              dc.l       QBRAN,SCOM2
              dc.l       AT,DOLIT,$80000000,ANDD       *?immediate
              dc.l       QBRAN,SCOM1
              dc.l       EXECU,EXIT              *its immediate, execute
SCOM1         dc.l       COMMA,EXIT              *its not immediate, compile
SCOM2         dc.l       TNUMB,ATEXE              *try to convert to number
              dc.l       QBRAN,SCOM3
              dc.l       LITER,EXIT              *compile number as integer
SCOM3         dc.l       THROW                     *error
 
*   OVERT       ( -- )
*              Link a new word into the current vocabulary.

              CODEM      5,'OVERT',OVERT
OVERT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       LAST,AT,CRRNT,AT,STORE,EXIT
 
*   *              ( -- )
*              Terminate a colon definition.

              CODEM     IMEDD+COMPO+1,';',SEMIS
SEMIS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       COMPI,EXIT,LBRAC,OVERT,EXIT
 
*   ]              ( -- )
*              Start compiling the words in the input stream.

              CODEM     1,']',RBRAC
RBRAC
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,SCOMP,TEVAL,STORE,EXIT
*   call,       ( ca -- )
*              Assemble a call instruction to ca.

*              COLONM     5,'call,'
          dc.l       CALLC,LINK_                     *;token pointer and link
LINK_     SET       *                            *;link points to a name string
          dc.b      5,'call',44               *;name string
          ALIGNM
CALLC
          nop
          jsr       DOLST
              dc.l       DOLIT,$4E714EB9,COMMA
              dc.l       COMMA,EXIT
 
*   :              ( -- * <string> )
*              Start a new colon definition using next word as its name.

              CODEM     1,':',COLON
COLON
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOKEN,SNAME,DOLIT,DOLST
              dc.l       CALLC,RBRAC,EXIT
 
*   IMMEDIATE       ( -- )
*              Make the last compiled word an immediate word.

              CODEM       9,'IMMEDIATE',IMMED
IMMED
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,IMEDD,LAST,AT,AT,ORR
              dc.l       LAST,AT,STORE,EXIT
 
*; Defining words
 
*   USER       ( u -- * <string> )
*              Compile a new user variable.

              CODEM     4,'USER',USER
USER
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOKEN,SNAME,OVERT
              dc.l       DOLIT,DOLST,CALLC
              dc.l       COMPI,DOUSE,COMMA,EXIT
 
*   CREATE       ( -- * <string> )
*              Compile a new array entry without allocating code space.

              CODEM       6,'CREATE',CREAT
CREAT
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOKEN,SNAME,OVERT
              dc.l       DOLIT,DOLST,CALLC
              dc.l       COMPI,DOVAR,EXIT
 
*   VARIABLE       ( -- * <string> )
*              Compile a new variable initialized to 0.

              CODEM       8,'VARIABLE',VARIA
VARIA
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       CREAT,DOLIT,0,COMMA,EXIT
 
*; Tools
 
*   _TYPE       ( b u -- )
*              Display a string. Filter non-printing characters.

              CODEM       5,'_TYPE',UTYPE
UTYPE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TOR                     *start count down loop
              dc.l       BRAN,UTYP2              *skip first pass
UTYP1         dc.l       DUPP,CAT,TCHAR,EMIT       *display only printable
              dc.l       DOLIT,1,PLUS              *increment address
UTYP2         dc.l       DONXT,UTYP1              *loop till done
              dc.l       DROP,EXIT
 
*   dm+              ( a u -- a )
*              Dump u bytes from , leaving a+u on the stack.

              CODEM       3,'dm+',DMP
DMP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       OVER,DOLIT,4,UDOTR       *display address
              dc.l       SPACE,TOR              *start count down loop
              dc.l       BRAN,PDUM2              *skip first pass
PDUM1         dc.l       DUPP,CAT,DOLIT,3,UDOTR       *display numeric data
              dc.l       DOLIT,1,PLUS              *increment address
PDUM2         dc.l       DONXT,PDUM1              *loop till done
              dc.l       EXIT
 
*   DUMP       ( a u -- )
*              Dump u bytes from a, in a formatted manner.

              CODEM       4,'DUMP',DUMP
DUMP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       BASE,AT,TOR,HEX              *save radix, set hex
              dc.l       DOLIT,16,SLASH              *change count to lines
              dc.l       TOR                     *start count down loop
DUMP1         dc.l       CR,DOLIT,16,DDUP,DMP       *display numeric
              dc.l       ROT,ROT
              dc.l       SPACE,SPACE,UTYPE       *display printable characters
              dc.l       NUFQ,INVER              *user control
              dc.l       QBRAN,DUMP2
              dc.l       DONXT,DUMP1              *loop till done
              dc.l       BRAN,DUMP3
DUMP2         dc.l       RFROM,DROP              *cleanup loop stack, early exit
DUMP3         dc.l       DROP,RFROM,BASE,STORE       *restore radix
              dc.l       EXIT
 
*   .S              ( ... -- ... )
*              Display the contents of the data stack.

              CODEM       2,'.S',DOTS
DOTS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       CR,DEPTH              *stack depth
              dc.l       TOR                     *start count down loop
              dc.l       BRAN,DOTS2              *skip first pass
DOTS1         dc.l       RAT,PICK,DOT              *index stack, display contents
DOTS2         dc.l       DONXT,DOTS1              *loop till done
*          DMM       DOTQP,' <sp'
          dc.l      DOTQP
          dc.b      4,' <sp'
          ALIGNM
              dc.l       EXIT
 
*   !CSP       ( -- )
*              Save stack pointer in CSP for error checking.

              CODEM       4,'!CSP',STCSP
STCSP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       SPAT,CSP,STORE,EXIT       *save pointer
 
*   ?CSP       ( -- )
*              Abort if stack pointer differs from that saved in CSP.

              CODEM       4,'?CSP',QCSP
QCSP
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       SPAT,CSP,AT,XORR       *compare pointers
*          DMM       ABORQ,'stacks'      *abort if different
          dc.l      ABORQ
          dc.b      6,'stacks'
          ALIGNM
              dc.l       EXIT
 
*   >NAME       ( ca -- na | F )
*              Convert code address to a name address.

              CODEM       5,'>NAME',TNAME
TNAME
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
          dc.l      CNTXT,AT                  *vocabulary link
TNAM1     dc.l      AT,QDUP
          dc.l      QBRAN,TNAM2
          dc.l      DDUP,NAMET,XORR
          dc.l      QBRAN,TNAM3
          dc.l      CELLM
          dc.l      BRAN,TNAM1
TNAM3     dc.l      ESWAP,DROP,EXIT
TNAM2     dc.l      DROP,DOLIT,0,EXIT

*   .ID              ( na -- )
*              Display the name at address.

              CODEM       3,'.ID',DOTID
DOTID
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       QDUP                     *if zero no name
              dc.l       QBRAN,DOTI1
              dc.l       COUNT,DOLIT,$1F,ANDD       *mask lexicon bits
              dc.l       UTYPE,EXIT              *display name string
DOTI1     dc.l      DOTQP
          dc.b      10,' {noName} '
          ALIGNM
* DMM       DOTQP,' {noName} '
              dc.l       EXIT
 
*   SEE              ( -- * <string> )
*              A simple decompiler.

              CODEM       3,'SEE',SEE
SEE
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       TICK                     *starting address
              dc.l       CR,CELLP
SEE1          dc.l       CELLP,DUPP,AT,DUPP       *?does it contain a zero
              dc.l       QBRAN,SEE2
              dc.l       TNAME                     *?is it a name
SEE2          dc.l       QDUP                     *name address or zero
              dc.l       QBRAN,SEE3
              dc.l       SPACE,DOTID              *display name
              dc.l       BRAN,SEE4
SEE3          dc.l       DUPP,AT,UDOT              *display number
SEE4          dc.l       NUFQ                     *user control
              dc.l       QBRAN,SEE1
              dc.l       DROP,EXIT
 
*   WORDS       ( -- )
*              Display the names in the context vocabulary.

              CODEM       5,'WORDS',WORDS
WORDS
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       CR,CNTXT,AT              *only in context
WORS1         dc.l       AT,QDUP
              dc.l       QBRAN,WORS2
              dc.l       DUPP,SPACE,DOTID       *display a name
              dc.l       CELLM,NUFQ              *user control
              dc.l       QBRAN,WORS1
              dc.l       DROP
WORS2         dc.l       EXIT
*
* FPACKS
*
          CODEM     6,'QPACKS',QPACKS
QPACKS
          move.l    (a3)+,a5
          move.l    a5,a0
          move.l    (a3)+,d2
          move.l    (a3)+,a1
          move.w    d2,(a5)+
FPACK1
          move.b    (a1)+,(a5)+
          dbf       d2,FPACK1
          move.l    a0,-(a3)
          NEXTM
*
          CODEM     3,'#IN',CIN
CIN
          move.l    a7,-(a3)
          NEXTM
*
          CODEM     4,'#OUT',COUT
COUT
          move.l    #CANAL,-(a3)
          NEXTM
*
          CODEM     8,'#DEFAULT',CDEF
CDEF
          move.l    #CDEFECTO,-(a3)
          NEXTM
*
LOAD2     move.l    (a3)+,a0
          move.l    #-1,d1
          move.l    #0,d3
          moveq     #1,d0
          trap      #2
          tst       d0
          bne       LOAD3
          move.l    a0,-(a7)
LOAD3
          NEXTM
*
* LOAD_FILE
*
          CODEM     9,'LOAD_FILE',LOAD
LOAD
          nop
          jsr       DOLST
          dc.l      BLANK,PARSE,HERE,QPACKS,LOAD2
          dc.l      EXIT

*
          CODEM     5,'TRAP1',TRAP1
TRAP1
          movem.l   a2/a3/a4,-(a7)
          move.l    (a3),a3
          move.l    (a3)+,d2
          move.l    (a3)+,d1
          move.l    (a3)+,d0
          trap      #1
          movem.l   (a7)+,a2/a3/a4
          adda.l    #12,a3
          move.l    d2,-(a3)
          move.l    d1,-(a3)
          move.l    d0,-(a3)
          NEXTM
*
          CODEM     5,'TRAP2',TRAP2
TRAP2
          move.l    (a3)+,a0
          move.l    (a3)+,d3
          move.l    (a3)+,d0
          trap      #2
          move.l    a0,-(a3)
          move.l    d0,-(a3)
          NEXTM
*
          CODEM     5,'TRAP3',TRAP3
TRAP3_
          move.l    (a3)+,d1
TRAP3
          move.l    (a3)+,a1
          move.l    (a3)+,a0
          move.l    (a3)+,d2
          move.l    (a3)+,d0
          moveq     #-1,d3
          trap      #3
          move.l    a1,-(a3)
          move.l    d1,-(a3)
          move.l    d0,-(a3)
          NEXTM
*
          CODEM     6,'TRAP3*',TRAP3_

*
          CODEM     6,'VEC_UT',VEC_UT
VEC_UT
          movem.l   a2/a3/a4,-(a7)
          move.l    (a3)+,a1
          move.l    (a3)+,a0
          move.l    (a3)+,d1
          move.l    (a3)+,a5
          jsr       (a5)
          movem.l   (a7)+,a2/a3/a4
          adda.l    #16,a3
          move.l    d3,-(a3)
          move.l    d0,-(a3)
          NEXTM

*; Hardware reset
 
*   VER              ( -- n )
*              Return the version number of this implementation.

              CODEM       3,'VER',VERSN
VERSN
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       DOLIT,VER*256+EXTE,EXIT
 
*   hi              ( -- )
*              Display the sign-on message of eForth.

              CODEM       2,'hi',HI
HI
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
              dc.l       STOIO,CR              *initialize I/O
*          DMM       DOTQP,'eForth v'    *model
          dc.l      DOTQP
          dc.b      8,'eFORTH v'
          ALIGNM
              dc.l       BASE,AT,HEX              *save radix
              dc.l       VERSN,BDIGS,DIG,DIG
              dc.l       DOLIT,46,HOLD
              dc.l       DIGS,EDIGS,TYPEE       *format version number
              dc.l       BASE,STORE,CR,EXIT       *restore radix
 
*   'BOOT       ( -- a )
*              The application startup vector.

*             COLONM     5,"'BOOT"
          dc.l       TBOOT,LINK_                     *;token pointer and link
LINK_     SET       *                            *;link points to a name string
          dc.b      5,39,'BOOT'               *;name string
          ALIGNM
TBOOT
          nop
          jsr       DOLST
              dc.l       DOVAR
              dc.l       HI                     *application to boot
 
*   COLD       ( -- )
*              The hilevel cold start sequence.
NAME_
          dc.l      COLD,LINK_
          dc.b      4,'COLD'
          ALIGNM
COLD
          nop                               *;align to cell boundary
          jsr       DOLST                            *;include CALL doLIST
COLD1         dc.l       DOLIT,UZERO,DOLIT,UPP
              dc.l       DOLIT,ULAST-UZERO,CMOVE       *initialize user area
              dc.l       PRESE                     *initialize stack and TIB
              dc.l       TBOOT,ATEXE              *application boot
              dc.l       LAST,CNTXT,STORE
              dc.l       CRRNT,CELLP,CRRNT,STORE
              dc.l       QUIT                     *start interpretation
              dc.l       BRAN,COLD1              *just in case

LASTN     equ       NAME_+8
NTOP      equ       NAMEE
CTOP      equ       *                          *next available memory in code dictionary
          END

